"""Top-level package for github-release-tool."""

from .cli import cli
from .release import Release
from .version import __author__, __email__, __timestamp__, __version__

__all__ = ["cli", "Release", __version__, __timestamp__, __author__, __email__]



#!/usr/bin/env python3

# vi: ft=python

import json
import os
import re
from pathlib import Path
from subprocess import check_output

import github3

VERSION_PATTERN = r"^([0-9]+)\.([0-9]+)\.([0-9]+)(-.*){0,1}$"
WHEEL_PATTERN = r"^([a-z][a-z0-9_]+)-([0-9]+\.[0-9]+\.[0-9]+)-.+\.whl$"
JSON_PATTERN = r"^([a-z][a-z0-9_]+)-([0-9]+\.[0-9]+\.[0-9]+)-release\.json$"


class Release:
    def __init__(
        self,
        *,
        organization=None,
        repository=None,
        token=None,
        module_dir=None,
        wheel_dir=None,
        version=None,
        local=False,
    ):
        organization = organization or os.environ["GITHUB_ORGANIZATION"]
        repository = repository or os.environ.get(
            "GITHUB_REPO", Path(".").resolve().stem
        )
        token = token or os.environ["GITHUB_TOKEN"]
        self.version_pattern = re.compile(VERSION_PATTERN)
        self.wheel_pattern = re.compile(WHEEL_PATTERN)
        self.json_pattern = re.compile(JSON_PATTERN)
        self.gh = github3.GitHub(token=token)
        if not isinstance(self.gh, github3.GitHub):
            raise RuntimeError("token login failed")
        self.repo = self.gh.repository(organization, repository)
        if not isinstance(self.repo, github3.repos.repo.Repository):
            raise RuntimeError(
                f"repo lookup failed: {organization}/{repository}"
            )
        if module_dir:
            self.module_dir = Path(module_dir).resolve()
            if not (self.module_dir / "__init__.py").is_file():
                raise RuntimeError(
                    f"module_dir '{str(module_dir)}'"
                    " does not appear to be a python module"
                )
        else:
            self.module_dir = None
        self.wheel_dir = wheel_dir or Path("./dist").resolve()
        self.local = local
        if version in [None, "latest"]:
            self.version = self.latest_release_version()
        else:
            self.version = self._check_version(version)

    def _check_version(self, v, return_none=False):
        if v is not None:
            if v.startswith("v"):
                v = v[1:]
            if not self.version_pattern.match(v):
                if return_none:
                    return None
                raise SyntaxError(f"unrecognized version format '{v}'")
        if not isinstance(v, str):
            if return_none:
                return None
            raise TypeError(f"expected string-type, got '{v}'")
        return v

    def local_release_files(self, wheel=False):
        """return dict of json or wheel files from ./dist"""
        if wheel:
            pattern = self.wheel_pattern
        else:
            pattern = self.json_pattern

        self.wheel_dir = self.wheel_dir or Path("./dist").resolve()

        if not self.wheel_dir.is_dir():
            raise RuntimeError(
                f"WHEEL_DIR {str(self.wheel_dir)} is not a directory"
            )

        ret = {}
        for _file in [e for e in self.wheel_dir.iterdir() if e.is_file()]:
            match = pattern.match(_file.name)
            if match:
                module, version = match.groups()[:2]
                if module == self.module_dir.name:
                    ret[self._check_version(version)] = _file
        return ret

    def _sort_versions(self, versions):
        """sort a list of semver strings"""
        key = {}
        for v in versions:
            code = "".join([f"{int(s):08x}" for s in v.split(".")])
            key[code] = v
        ret = [key[code] for code in sorted(key.keys())]
        return list(ret)

    def latest_release_version(self, local=False):
        ret = None
        if self.local or local:
            versions = self.local_release_versions()
            if len(versions) > 0:
                versions = self._sort_versions(versions)
                ret = versions[-1]
        else:
            releases = self._validated_releases(self.repo.releases())
            if len(list(releases)):
                versions = self._sort_versions(
                    [self._check_version(r.tag_name) for r in releases]
                )
                if len(versions):
                    ret = versions[-1]
        if ret:
            ret = self._check_version(ret)
        return ret

    def _validated_releases(self, releases):
        ret = []
        for r in list(releases):
            if self._check_version(r.tag_name, return_none=True):
                ret.append(r)
        return ret

    def get_release_data(self):
        v = self.version
        if self.local:
            releases = self.local_release_files()
            if v in releases:
                _file = releases[v]
                return json.loads(Path(_file).read_text())
        else:
            return self._get_repo_release().as_dict()
        return None

    def local_release_versions(self):
        """return list of versions of local wheel files"""
        wheels = self.local_release_files(wheel=True)
        versions = list(wheels.keys())
        return versions

    def list_release_versions(self, sorted=True):
        if self.local:
            ret = self.local_release_versions()
        else:
            ret = [
                self._check_version(r.tag_name)
                for r in self._validated_releases(self.repo.releases())
            ]

        if sorted:
            ret = self._sort_versions(ret)

        return ret

    def get_current_branch(self):
        out = check_output("git branch | awk '/^\\*/{print $2}'", shell=True)
        branch = out.decode().strip()
        return branch

    def create_release(self, **kwargs):
        """create a new release"""

        ret = None

        if "tag_name" in kwargs:
            version = self._check_version(kwargs["tag_name"])
            kwargs.pop("version", None)
        else:
            version = kwargs.pop(
                "version", self.latest_release_version(local=True)
            )

        kwargs.setdefault("tag_name", f"v{version}")
        kwargs.setdefault("target_commitish", self.get_current_branch())
        kwargs.setdefault("name", f"v{version}")
        kwargs.setdefault("body", f"Release of version v{version}")
        kwargs.setdefault("draft", False)
        kwargs.setdefault("prerelease", False)

        verify = kwargs.pop("verify", None)
        if not verify or verify(kwargs):
            release = self.repo.create_release(**kwargs)
            if release:
                ret = release.as_dict()
        return ret

    def _get_wheel(self):
        wheel = self.local_release_files(wheel=True)[self.version]
        wheel = Path(wheel).resolve()
        return wheel

    def _get_repo_release(self):
        """return current remote release"""
        release = self.repo.release_from_tag(f"v{self.version}")
        if not release:
            RuntimeError(f"unknown release: {self.version}")
        return release

    def upload_asset(
        self, asset=None, content_type=None, label=None, verify=None
    ):
        """upload a binary asset to release"""
        release = self._get_repo_release()

        if not asset:
            asset = self._get_wheel()
        asset = Path(asset).resolve()

        content_type = content_type or "application/binary"
        name = asset.name

        if verify:
            verify(
                dict(
                    release=release.tag_name,
                    content_type=content_type,
                    label=label,
                    name=name,
                    local_file=str(asset),
                )
            )

        with asset.open("rb") as ifp:
            response = release.upload_asset(
                content_type=content_type,
                name=name,
                asset=ifp,
                label=label,
            )
            if response:
                return response.as_dict()

        return None

    def get_assets(self):
        """return the assets from the selected remote release"""
        release = self._get_repo_release()
        return [asset.as_dict() for asset in release.assets()]

    def download_assets(
        self, _id=None, regex=None, path=Path("."), dry_run=False, update=False
    ):
        """download the assets, filter name by regex, optionally deleting old versions"""
        ret = []
        path = Path(path).resolve()
        release = self._get_repo_release()
        for asset in release.assets():
            if _id and asset.id != _id:
                continue
            elif regex and not re.match(regex, asset.name):
                continue
            asset_path = path / asset.name
            if dry_run:
                result = asset_path
            else:
                if update:
                    self.delete_old_versions(asset, path)
                result = asset.download(asset_path)
            ret.append(str(result))

        return ret

    def delete_old_versions(self, asset, path):
        """delete old versions of the asset"""
        suffix = Path(asset.name).suffix
        basename = asset.name.split("-")[0]
        files = filter(lambda f: f.is_file(), path.iterdir())
        asset_files = filter(lambda f: f.suffix == suffix, files)
        for asset_file in asset_files:
            file_base, file_version = asset_file.name.split("-")[:2]
            if file_base == basename:
                if file_version != self.version:
                    asset_file.unlink()

    def download_file(self, repo_path, output_file):
        """download the contents of a repo file and write to output_file"""
        release = self._get_repo_release()
        ref = release.tag_name
        output_file.write(self.repo.file_contents(repo_path, ref).decoded)
        output_file.close()
        return 0

    def wheel(self):
        """return the filename of the selected wheel"""
        return str(self._get_wheel())



"""Console script for github_release_tool."""

import json
import sys
from pathlib import Path

import click

from .release import Release
from .version import __timestamp__, __version__

header = f"{__name__.split('.')[0]} v{__version__} {__timestamp__}"


def output_setup(_json=True, _compact=False, _func=print):
    def _output(data):
        if _json:
            if _compact:
                args = dict(separators=(",", ":"))
            else:
                args = dict(indent=2)
            ret = json.dumps(data, **args)
        else:
            ret = str(data)

        _func(ret)
        return 0

    return _output


class CustomGroup(click.Group):
    def get_help(self, *args, **kwargs):
        help_str = super().get_help(*args, **kwargs)
        help_list = help_str.split("\n")
        help_list[2] += f" v{__version__}"
        return "\n".join(help_list)


@click.group(name="release", cls=CustomGroup)
@click.option("-d", "--debug", is_flag=True, help="debug mode")
@click.option(
    "-j/-J",
    "--json/--no-json",
    is_flag=True,
    default=True,
    help="format output as json",
)
@click.option(
    "-c",
    "--compact",
    is_flag=True,
    help="compact json output",
)
@click.option(
    "-v",
    "--version",
    type=str,
    help="select release version tag",
)
@click.option(
    "-t", "--token", type=str, envvar="GITHUB_TOKEN", show_envvar=True
)
@click.option(
    "-o",
    "--organization",
    type=str,
    envvar="GITHUB_ORG",
    show_envvar=True,
    help="github organization name",
)
@click.option(
    "-r",
    "--repository",
    type=str,
    envvar="GITHUB_REPO",
    show_envvar=True,
    help="github repository name",
)
@click.option(
    "-m",
    "--module-dir",
    type=click.Path(file_okay=False, writable=True, path_type=Path),
    envvar="MODULE_DIR",
    show_envvar=True,
    help="python module directory",
)
@click.option(
    "-w",
    "--wheel-dir",
    type=click.Path(
        dir_okay=True, file_okay=False, writable=True, path_type=Path
    ),
    envvar="WHEEL_DIR",
    show_envvar=True,
    help="wheel directory (usually ./dist)",
)
@click.option(
    "-l",
    "--local",
    is_flag=True,
    help="select local release data",
)
@click.pass_context
def cli(ctx, debug, json, compact, **kwargs):
    """github release tool"""

    if kwargs["local"]:
        if kwargs["module_dir"] is None or not kwargs["module_dir"].is_dir():
            ctx.fail("MODULE_DIR is not a directory")

    def exception_handler(
        exception_type, exception, traceback, debug_hook=sys.excepthook
    ):
        if debug:
            debug_hook(exception_type, exception, traceback)
        else:
            click.echo(f"{exception_type.__name__}: {exception}", err=True)
            sys.exit(-1)

    sys.excepthook = exception_handler

    ctx.obj = Release(**kwargs)
    ctx.obj.output = output_setup(json, compact, click.echo)


@cli.command()
@click.pass_context
def list(ctx):
    """list all released or dist versions"""
    r = ctx.obj
    return r.output(r.list_release_versions())


@cli.command()
@click.pass_context
def latest(ctx):
    """latest release"""
    r = ctx.obj
    return r.output(r.latest_release_version())


@cli.command()
@click.pass_context
def assets(ctx):
    """list asset data for selected version or latest release"""
    r = ctx.obj
    return r.output(r.get_assets())


@cli.command()
@click.pass_context
@click.option("-i", "--id", "_id", type=int, help="identify asset by id")
@click.option("-r", "--regex", type=str, help="filter asset filenames")
@click.option(
    "-d", "--dry-run", is_flag=True, help="simulate action and report"
)
@click.option("-u", "--update", is_flag=True, help="delete old versions")
@click.argument(
    "path",
    type=click.Path(
        exists=True, file_okay=False, writable=True, path_type=Path
    ),
)
def download_asset(ctx, _id, regex, path, dry_run, update):
    """download asset with id to path"""
    r = ctx.obj
    return r.output(r.download_assets(_id, regex, path, dry_run, update))


@cli.command()
@click.argument("repo-path", type=str)
@click.argument("output", type=click.File("wb"), default="-")
@click.pass_context
def download_file(ctx, repo_path, output):
    """output the contents of a repo file"""
    r = ctx.obj
    return r.download_file(repo_path, output)


@cli.command()
@click.pass_context
def wheel(ctx):
    """output wheel pathname"""
    r = ctx.obj
    return r.output(r.wheel())


def verify_upload(args):
    click.echo("Upload asset file:", err=True)
    for k, v in args.items():
        click.echo(f"{k}={v}")
    click.confirm("confirm", abort=True, err=True)
    return True


@cli.command()
@click.option(
    "-c",
    "--content-type",
    type=str,
    help="content-type",
)
@click.option("-l", "--label", type=str, help="short description")
@click.option("-f", "--force", is_flag=True, help="bypass confirmation prompt")
@click.argument(
    "asset",
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
    default=None,
    required=False,
)
@click.pass_context
def upload(ctx, content_type, label, force, asset):
    """upload asset file to release"""
    r = ctx.obj
    if force:
        verify = None
    else:
        verify = verify_upload
    return r.output(r.upload_asset(asset, content_type, label, verify))


@cli.command()
@click.argument("key", type=str, default=None, required=False)
@click.pass_context
def get(ctx, key):
    """output release response data"""
    r = ctx.obj

    data = r.get_release_data()

    if not None and key is not None:
        data = data.get(key, None)

    return r.output(data)


def verify_create(args):
    click.echo("create release:", err=True)
    for k, v in args.items():
        click.echo(f"{k}={v}", err=True)
    click.confirm("confirm", abort=True, err=True)
    return True


@cli.command()
@click.option("-t", "--tag-name", type=str, help="tag name")
@click.option(
    "-c",
    "--target-commitish",
    type=str,
    help="target commitish (branch or commit)",
)
@click.option("-n", "--name", type=str, help="release name")
@click.option("-b", "--body", type=str, help="body text for release")
@click.option("-d", "--draft", is_flag=True, help="draft mode switch")
@click.option(
    "-p", "--prerelease", is_flag=True, help="prerelease mode switch"
)
@click.option("-f", "--force", is_flag=True, help="bypass confirmation prompt")
@click.pass_context
def create(ctx, **kwargs):
    """create a new release"""
    r = ctx.obj
    if not kwargs.pop("force", False):
        kwargs["verify"] = verify_create
    kwargs = {k: v for k, v in kwargs.items() if v}
    return r.output(r.create_release(**kwargs))


if __name__ == "__main__":
    sys.exit(cli())  # pragma: no cover



#!/usr/bin/env python

"""Tests for `github_release_tool` package."""

import os
from logging import debug

import pytest
from click.testing import CliRunner

import github_release_tool
from github_release_tool import cli


@pytest.fixture
def runner():
    runner = CliRunner()

    def _runner(*args, **kwargs):
        _args = tuple([cli] + list(args))
        debug(f"cli({_args[1:]}, {kwargs})")
        result = runner.invoke(*_args, **kwargs)
        assert result.exit_code == 0, result
        debug(result.output)
        debug(f"{result}")
        return result

    return _runner


def test_version():
    """Test reading version and module name"""
    assert github_release_tool.__name__ == "github_release_tool"
    assert isinstance(github_release_tool.__version__, str)


def test_cli():
    """Test the CLI."""
    runner = CliRunner()
    result = runner.invoke(cli)
    assert result.exit_code == 0, result

    result = runner.invoke(cli, ["--bad-option"])
    assert result.exception, result
    assert result.exit_code != 0, result

    result = runner.invoke(cli, ["--help"])
    assert result.exit_code == 0, result
    assert "Show this message and exit." in result.output, result


def test_cli_latest_remote():
    os.environ.pop("MODULE_DIR", None)
    runner = CliRunner()
    result = runner.invoke(cli, ["latest"], catch_exceptions=False)
    assert not result.exception
    assert result.exit_code == 0, result


def test_cli_latest_local():
    os.environ.pop("MODULE_DIR", None)
    runner = CliRunner()
    result = runner.invoke(cli, ["-l", "latest"])
    assert result.exception
    assert "MODULE_DIR is not a directory" in result.output



import os
from logging import debug, info
from pathlib import Path
from pprint import pformat

import pytest

from github_release_tool import Release


@pytest.fixture
def module_dir():
    return Path(".").resolve() / "github_release_tool"


@pytest.fixture
def wheel_dir():
    ret = Path(".").resolve() / "dist"
    ret.mkdir(exist_ok=True)
    return ret


@pytest.fixture
def args(module_dir, wheel_dir):
    return dict(
        organization="rstms",
        repository="github-release-tool",
        token=os.environ["GITHUB_TOKEN"],
        module_dir=module_dir,
        wheel_dir=wheel_dir,
    )


@pytest.fixture
def mkrelease(args):
    def _release(**kwargs):
        args.update(kwargs)
        r = Release(**args)
        assert r
        debug(r)
        assert r.repo
        debug(r.repo)
        return r

    return _release


def test_release_list_local(mkrelease):
    r = mkrelease(local=True)
    ret = r.list_release_versions()
    assert isinstance(ret, list)
    assert len(ret)
    assert isinstance(ret[0], str)
    info(ret)


def test_release_list_remote(mkrelease):
    r = mkrelease()
    ret = r.list_release_versions()
    assert isinstance(ret, list)
    assert len(ret)
    assert isinstance(ret[0], str)
    info(ret)


def test_release_latest_local(mkrelease):
    r = mkrelease(local=True)
    ret = r.latest_release_version()
    assert isinstance(ret, str)
    info(ret)


def test_release_latest_remote(mkrelease):
    r = mkrelease()
    ret = r.latest_release_version()
    assert isinstance(ret, str)
    info(ret)


def test_release_sort(mkrelease):
    r = mkrelease()
    before = [
        "0.2.4",
        "11.3.1",
        "4.22.1",
        "1.2.2",
        "9.1.22",
        "1.4.2",
        "3.2.3",
        "4.2.1",
    ]
    debug(pformat({"before": before}))
    after = r._sort_versions(before)
    debug(pformat({"after": after}))
    assert after[-1] == "11.3.1"


def test_release_get(mkrelease):
    r = mkrelease()
    ret = r.get_release_data()
    assert ret
    assert isinstance(ret, dict)
    info(pformat(ret))


TEST_VERSION = "0.0.6"


def test_release_check_version(mkrelease):
    r = mkrelease()

    ret1 = r._check_version(TEST_VERSION)
    assert ret1

    with pytest.raises(TypeError):
        r._check_version(None)

    with pytest.raises(SyntaxError):
        r._check_version("foo")

    ret4 = r._check_version("0.0.0")
    assert ret4


def test_release_local_release_files_wheel(mkrelease):
    r = mkrelease()
    ret = r.local_release_files(wheel=True)
    assert isinstance(ret, dict)
    info(ret)


def test_release_local_release_files_json(mkrelease):
    r = mkrelease()
    ret = r.local_release_files()
    assert isinstance(ret, dict)
    info(ret)


def test_release_wheel(mkrelease):
    r = mkrelease()
    ret = r.wheel()
    assert isinstance(ret, str)
    info(ret)


def test_release_get_current_branch(mkrelease):
    r = mkrelease()
    ret = r.get_current_branch()
    assert isinstance(ret, str)
    assert ret == "master"


def test_release_get_assets(mkrelease):
    r = mkrelease()
    ret = r.get_assets()
    assert isinstance(ret, list)
    info(ret)


def test_release_local_release_versions(mkrelease):
    r = mkrelease()
    ret = r.local_release_versions()
    assert isinstance(ret, list)
    assert ret
    assert isinstance(ret[0], str)
    info(ret)


def test_release_download_assets(mkrelease, shared_datadir):
    r = mkrelease()
    dist = shared_datadir / "dist"
    dist.mkdir()
    ret = r.download_assets(regex=r".*.whl", path=dist)
    assert isinstance(ret, list)
    assert len(ret) == 1
    assert len(list(dist.iterdir())) == 1
    assert isinstance(ret[0], str)
    assert ret[0].endswith(".whl")


def test_release_update_assets(mkrelease, shared_datadir):
    release_versions = mkrelease().list_release_versions()
    dist = shared_datadir / "dist"
    dist.mkdir()
    for version in release_versions[-4:]:
        r = mkrelease(version=version)
        ret = r.download_assets(regex=r".*.whl", path=dist)
        assert isinstance(ret, list)
        assert len(ret) == 1
    assert len(list(dist.iterdir())) == 4

    r = mkrelease()
    ret = r.download_assets(regex=r".*.whl", path=dist, update=True)
    assert isinstance(ret, list)
    assert len(ret) == 1
    assert isinstance(ret[0], str)
    assert ret[0].endswith(".whl")
    assert len(list(dist.iterdir())) == 1


"""


    def create_release(self, **kwargs):
    def _get_repo_release(self):
    def upload_asset(
"""
